package org.drupal.project.recommender;

import org.apache.mahout.cf.taste.common.TasteException;
import org.apache.mahout.cf.taste.impl.common.LongPrimitiveIterator;
import org.apache.mahout.cf.taste.impl.model.jdbc.*;
import org.apache.mahout.cf.taste.impl.recommender.GenericBooleanPrefItemBasedRecommender;
import org.apache.mahout.cf.taste.impl.recommender.GenericItemBasedRecommender;
import org.apache.mahout.cf.taste.impl.similarity.*;
import org.apache.mahout.cf.taste.model.DataModel;
import org.apache.mahout.cf.taste.model.JDBCDataModel;
import org.apache.mahout.cf.taste.recommender.ItemBasedRecommender;
import org.apache.mahout.cf.taste.recommender.RecommendedItem;
import org.apache.mahout.cf.taste.recommender.Recommender;
import org.apache.mahout.cf.taste.similarity.ItemSimilarity;
import org.apache.mahout.cf.taste.similarity.UserSimilarity;
import org.drupal.project.computing.*;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.concurrent.*;

/**
 * Implementation of classical collaborative filtering algorithms that uses the user-item preference
 * relationship from Mahout. This class is written as if it's Item2Item.
 *
 * ODO list for refactor:
 * - File recommender is a hack. Use the OO way (implements FileRecommender)
 * - genericComputeSave() is too complicated.
 * - remove sequential execution and use Thread alone.
 * - architectural support for Hadoop
 */
abstract public class MahoutRecommender extends RecommenderAlgorithm {

    /**
     * How many similarity/prediction records to keep for each item.
     */
    private final int DEFAULT_MAX_KEEP = 100;

    /**
     * when numbers of users/items lower than this number, load into memory first in "auto" mode.
     */
    private final int LOAD_MEMORY_THRESHOLD = 1000000;

    /**
     * how many records processed to show progress info in genericComputeSave()
     */
    private final int PROGRESS_INCREMENT = 500;

    // unless you understand what similarity is useful, just choose AUTO and delegate to the algorithm to decide.
    protected enum SimilarityType {
        AUTO, CITYBLOCK, EUCLIDEAN, LOGLIKELIHOOD, PEARSON, SPEARMAN, TANIMOTO, COSINE
    }

    protected enum PerformanceType {
        MEMORY, DATABASE, AUTO
    }

    protected enum PreferenceType {
        SCORE, BOOLEAN
    }


    protected String table;
    protected String fieldUser;
    protected String fieldItem;
    protected String fieldPreference;
    protected String fieldTimestamp;

    protected boolean stagingUseView;
    protected String viewName;

    protected SimilarityType similarityType;
    protected PerformanceType performanceType;
    protected PreferenceType preferenceType;

    protected int maxKeep;
    protected String predictionSourceFilter;
    protected String predictionScoreFilter;
    protected String similarityScoreFilter;


    protected DataModel dataModel;
    protected ItemSimilarity itemSimilarity;
    protected UserSimilarity userSimilarity;
    protected Recommender recommender;

    /**
     * Parse configuration.
     * @param params
     * @throws DConfigException
     */
    @Override
    public void parseConfig(Map<String, Object> params) throws DConfigException {
        parseTable(params);
        parseOptional(params);
    }

    private void parseTable(Map<String, Object> params) throws DConfigException {
        if (!params.containsKey("table")) {
            throw new DConfigException("You have to provide a SQL query, a table name, <FILE>, or <BUILTIN> for the user-item preference.");
        } else {
            table = (String) params.get("table");
        }

        // process <BUILTIN> and <FILE>, and leave SQL or Table name untouched.

        if (table.equals("<BUILTIN>")) {
            table = "SELECT source_id, target_id, score, updated FROM {recommender_preference} WHERE recommender = '" + recommenderName + "'";
            fieldUser = "source_id";
            fieldItem = "target_id";
            fieldPreference = "score";
            fieldTimestamp = "updated";
        }
        else if (table.equals("<FILE>")) {
            throw new UnsupportedOperationException("Currently not support <FILE> as table name.");
        }
        else {
            ArrayList<String> fields = new ArrayList<String>((List) params.get("fields"));
            // description of the fields see: org.apache.mahout.cf.taste.impl.model.file.FileDataModel
            switch (fields.size()) {
                case 4:
                    //appFieldTimestamp = appFields.get(3) == SerializedPhpParser.NULL ? null : appFields.get(3);
                    fieldTimestamp = fields.get(3); // TODO: verify NULL is the same null.
                case 3:
                    //appFieldPreference = appFields.get(2) == SerializedPhpParser.NULL ? null : appFields.get(2);
                    fieldPreference = fields.get(2);
                case 2:
                    fieldUser = fields.get(0);
                    fieldItem = fields.get(1);
                    break;
                default:
                    throw new DConfigException("Wrong parameters in recommender fields");
            }
        }

        // attention: callers need to make sure using view is OK.
        if (params.containsKey("staging") && ((String) params.get("staging")).trim().toLowerCase().equals("view")) {
            stagingUseView = true;
            viewName = "{recommender_preference_" + recommenderName + "}";
        } else {
            stagingUseView = false;
        }
    }


    private void parseOptional(Map<String, Object> params) throws DConfigException {
        if (params.containsKey("similarity")) {
            similarityType = SimilarityType.valueOf(((String) params.get("similarity")).toUpperCase());
        }

        performanceType = params.containsKey("performance") ?
                PerformanceType.valueOf(((String) params.get("performance")).toUpperCase()) : PerformanceType.AUTO;

        preferenceType = params.containsKey("preference") ?
                PreferenceType.valueOf(((String) params.get("preference")).toUpperCase()) : PreferenceType.BOOLEAN;
        if (preferenceType == PreferenceType.SCORE && fieldPreference == null) {
            throw new DConfigException("Please specify preference field if score is used.");
        }

        maxKeep = params.containsKey("max_keep") ? (Integer) params.get("max_keep") : DEFAULT_MAX_KEEP;

        if (params.containsKey("prediction_source_filter")) {
            predictionSourceFilter = (String) params.get("prediction_source_filter");
            if (!predictionSourceFilter.matches("<\\d+")) {
                logger.warning("Currently the program only supports prediction_source_filter <N");
                predictionSourceFilter = null;
            }
        }

        if (params.containsKey("prediction_score_filter")) {
            predictionScoreFilter = (String) params.get("prediction_score_filter");
            if (!predictionScoreFilter.matches(">[0-9.]+")) {
                logger.warning("Currently the program only supports prediction_score_filter >x");
                predictionScoreFilter = null;
            }
        }

        if (params.containsKey("similarity_score_filter")) {
            similarityScoreFilter = (String) params.get("similarity_score_filter");
            if (!similarityScoreFilter.matches(">[0-9.]+")) {
                logger.warning("Currently the program only supports similarity_score_filter >x");
                similarityScoreFilter = null;
            }
        }
    }



    /**
     * If the input is a SQL query, then staging the data into {recommender_preference_staging} first.
     * Or, create a database view if "staging" is set to view.
     */
    private void processTable() throws DConnectionException, DConfigException {

        if (table.matches("\\{.*\\}")) {
            table = database.d(table);  // escape the table.
            return;  // that means the app use existing table, and no need to process staging table.
        }

        if (!table.trim().toUpperCase().startsWith("SELECT")) {
            throw new DConfigException("The SQL query should starts with SELECT");
        }

        if (stagingUseView) {
            try {
                database.update("CREATE VIEW " + viewName + " AS " + table);
                logger.info("Using database view " + viewName + " as preference table.");
                table = database.d(viewName);
                return;
            } catch (Throwable e) {
                logger.warning("Cannot create view for recommender: " + recommenderName + ". Fall back to use recommender_preference_staging. Please set 'staging'='table'");
            }
        }

        logger.info("Using {recommender_preference_staging} table. Loading data.");

        // clean the table.
        database.update("DELETE FROM {recommender_preference_staging}");

        List<Map<String, Object>> results = database.query(table);
        Object[][] values = new Object[results.size()][4];
        int count = 0;
        for (Map<String, Object> row : results) {
            values[count][0] = row.get(fieldUser);
            values[count][1] = row.get(fieldItem);
            values[count][2] = fieldPreference != null ? row.get(fieldPreference) : null;
            values[count][3] = fieldTimestamp != null ? row.get(fieldTimestamp) : updatedTimestamp;
            count ++;
        }

        // TODO: we can save some insert data in "increment" mode. Also use batch uploader.
        database.batch("INSERT INTO {recommender_preference_staging}(source_id, target_id, score, updated) VALUES(?, ?, ?, ?)", values);
        //appSql = null;  // could save this variable if needed.
        table = database.d("{recommender_preference_staging}");
        fieldUser = "source_id";
        fieldItem = "target_id";
        fieldPreference = "score";
        fieldTimestamp = "updated";
    }

    @Override
    protected void beforeExecute() {
        super.beforeExecute();
        try {
            processTable();
        } catch (DConnectionException e) {
            throw new DRuntimeException("Cannot process table.", e);
        } catch (DConfigException e) {
            throw new DRuntimeException("Configuration error when processing table.", e);
        }
    }

    @Override
    protected void afterExecute() {
        super.afterExecute();
        if (stagingUseView) {
            try {
                database.update("DROP VIEW " + viewName);
            } catch (Throwable e) {
                logger.warning("Cannot drop view " + viewName + ". Please set staging=table instead.");
            }
        }
    }


    @Override
    protected void execute() throws DCommandExecutionException, DConnectionException {
        try {
            logger.info("Initializing data model, similarity and recommender.");
            initDataModel();
            numUsers = dataModel.getNumUsers();
            numItems = dataModel.getNumItems();
            //printMessage("Users: " + numUsers + ". Items: " + numItems + ".");

            initSimilarity();
            String className = itemSimilarity != null ? itemSimilarity.getClass().getName() : userSimilarity != null ? userSimilarity.getClass().getName() : "None";
            logger.info("Using similarity class: " + className);

            initRecommender();
            logger.info("Using recommender class: " + recommender.getClass().getName());

            logger.info("Computing and saving similarity data.");
            computeSimilarity();

            logger.info("Computing and saving prediction data.");
            computePrediction();

            //int timeSpent = (int) (DrupletUtils.getLocalUnixTimestamp() - updatedTimestamp);
            //message.append(" (Time spent: ").append(timeSpent/3600).append("h").append(timeSpent / 60 % 60).append("m").append(timeSpent%60).append("s)");
        } catch (TasteException e) {
            throw new DCommandExecutionException(e);
        } catch (SQLException e) {
            throw new DConnectionException(e);
        }
    }


    protected void initDataModel() throws TasteException {
        logger.info("Initializing data model.");
        // attention: we do the wrap here. but in fact we already used DBCP for pooling.
        ConnectionPoolDataSource wrapperDataSource = new ConnectionPoolDataSource(database.getDataSource());
        if (database.getDatabaseDriver().equals(DDatabase.DatabaseDriver.MYSQL) && preferenceType == PreferenceType.BOOLEAN) {
            dataModel = new MySQLBooleanPrefJDBCDataModel(wrapperDataSource, table, fieldUser, fieldItem, fieldTimestamp);
        }
        else if (database.getDatabaseDriver().equals(DDatabase.DatabaseDriver.MYSQL) && preferenceType == PreferenceType.SCORE) {
            dataModel = new MySQLJDBCDataModel(wrapperDataSource, table, fieldUser, fieldItem, fieldPreference, fieldTimestamp);
        }
        else if (database.getDatabaseDriver().equals(DDatabase.DatabaseDriver.PGSQL) && preferenceType == PreferenceType.BOOLEAN) {
            dataModel = new PostgreSQLBooleanPrefJDBCDataModel(wrapperDataSource, table, fieldUser, fieldItem, fieldTimestamp);
        }
        else if (database.getDatabaseDriver().equals(DDatabase.DatabaseDriver.PGSQL) && preferenceType == PreferenceType.SCORE) {
            dataModel = new PostgreSQLJDBCDataModel(wrapperDataSource, table, fieldUser, fieldItem, fieldPreference, fieldTimestamp);
        }
        else if (preferenceType == PreferenceType.BOOLEAN) {
            dataModel = new SQL92BooleanPrefJDBCDataModel(wrapperDataSource, table, fieldUser, fieldItem, fieldTimestamp);
        }
        else if (preferenceType == PreferenceType.SCORE) {
            dataModel = new SQL92JDBCDataModel(wrapperDataSource, table, fieldUser, fieldItem, fieldPreference, fieldTimestamp);
        }
        else {
            assert false;
        }

        if (performanceType == null || performanceType == PerformanceType.AUTO) {
            if (dataModel.getNumUsers() < LOAD_MEMORY_THRESHOLD && dataModel.getNumItems() < LOAD_MEMORY_THRESHOLD) {
                performanceType = PerformanceType.MEMORY;
            }
        }

        if (performanceType == PerformanceType.MEMORY) {
            logger.info("Switching to MEMORY mode. Load all data from database into memory first.");
            dataModel = new ReloadFromJDBCDataModel((JDBCDataModel) dataModel);
        }
    }

    protected void initSimilarity() throws TasteException {
        if (similarityType == null || similarityType == SimilarityType.AUTO) {
            // automatically set common similarity model based on whether it's boolean or score preference.
            if (preferenceType == PreferenceType.BOOLEAN) {
                itemSimilarity = new LogLikelihoodSimilarity(dataModel);
            } else if (preferenceType == PreferenceType.SCORE) {
                itemSimilarity = new PearsonCorrelationSimilarity(dataModel);
            } else {
                assert false;
            }
        } else if (similarityType == SimilarityType.CITYBLOCK) {
            itemSimilarity = new CityBlockSimilarity(dataModel);
        } else if (similarityType == SimilarityType.EUCLIDEAN) {
            itemSimilarity = new EuclideanDistanceSimilarity(dataModel);
        } else if (similarityType == SimilarityType.LOGLIKELIHOOD) {
            itemSimilarity = new LogLikelihoodSimilarity(dataModel);
        } else if (similarityType == SimilarityType.PEARSON) {
            itemSimilarity = new PearsonCorrelationSimilarity(dataModel);
        } else if (similarityType == SimilarityType.TANIMOTO) {
            itemSimilarity = new TanimotoCoefficientSimilarity(dataModel);
        } else if (similarityType == SimilarityType.COSINE) {
            itemSimilarity = new UncenteredCosineSimilarity(dataModel);
        } else if (similarityType == SimilarityType.SPEARMAN) {
            // no SpearmanCorrelationSimilarity for itemSimilarity
            userSimilarity = new SpearmanCorrelationSimilarity(dataModel);
            itemSimilarity = null;
        } else {
            assert false;  // no other possibilities.
        }

        // set userSimilarity to be the same as itemSimilarity.
        // indeed, these two are just different in names.
        if (userSimilarity == null && itemSimilarity != null) {
            userSimilarity = (UserSimilarity) itemSimilarity;
        }
    }

    protected void initRecommender() throws TasteException {
        if (preferenceType == PreferenceType.BOOLEAN) {
            recommender = new GenericBooleanPrefItemBasedRecommender(dataModel, itemSimilarity);
        } else if (preferenceType == PreferenceType.SCORE) {
            recommender = new GenericItemBasedRecommender(dataModel, itemSimilarity);
        } else {
            assert false;
        }
    }


    // Fixme: if we throw unhandled exception here, who takes care of rollback?
    // Fixme: dirty read? [#1185100]
    protected int genericComputeSave(EntityHandler handler, LongPrimitiveIterator entityIterator, int numEntity, String tableIdentifier, boolean rebuild) throws TasteException, DConnectionException, SQLException {
        DDatabase.AsyncBatchRunner deleteBatch = null;
        DDatabase.AsyncBatchRunner insertBatch = null;
        Connection connection = null;
        int threadNumber = ((DefaultApplication) getApplication()).getThreadNumber();

        // handle database, check database status.
        String tableName = "{recommender_" + tableIdentifier + "}";
        Long ts = DUtils.getInstance().getLong(database.queryValue("SELECT max(updated) FROM " + tableName));
        long pastMaxTimestamp = ts == null ? 0 : ts;

        if (pastMaxTimestamp > updatedTimestamp) {
            logger.severe("Please reinstall Recommender API or manually remove all data from " + tableName + ". It has records with timestamp newer than the current timestamp");
            throw new DRuntimeException("Internal database error. Please reinstall recommender module of your Drupal site.");
        }

        // set db connection
        connection = database.getConnection();
        connection.setAutoCommit(false);

        // #1428232: some database doesn't support this transaction level.
        if (connection.getMetaData().supportsTransactionIsolationLevel(Connection.TRANSACTION_READ_COMMITTED)) {
            connection.setTransactionIsolation(Connection.TRANSACTION_READ_COMMITTED);
        }

        // handle delete
        String deleteSql;
        if (rebuild) {
            logger.fine("Rebuilding " + tableName + " for recommender app with name: " + recommenderName);
            deleteSql = database.d("DELETE FROM " + tableName + " WHERE recommender = ? AND updated < ?");
        } else {
            deleteSql = database.d("DELETE FROM " + tableName + " WHERE recommender = ? AND source_id = ? AND updated < ?");
        }
        deleteBatch = new DDatabase.AsyncBatchRunner(connection, deleteSql, database.getMaxBatchSize());
        if (rebuild) {
            deleteBatch.put(recommenderName, updatedTimestamp);
        }
        deleteBatch.start();   // start the thread

        // handle insert
        String insertSql = database.d("INSERT INTO " + tableName + "(recommender, source_id, target_id, score, updated) VALUES(?, ?, ?, ?, ?)");
        insertBatch = new DDatabase.AsyncBatchRunner(connection, insertSql, database.getMaxBatchSize());
        insertBatch.start();


        int count = 0;
        // attention: this might be set by multi-threads, but we shouldn't have atomic problem.
        int resultCount = 0;

        // process data

//            // TODO: this is sequential execution, make sure the multi-thread version is correct.
//            logger.info("Start computing recommendation and saving results in serial.");
//
//            while (entityIterator.hasNext()) {
//                count ++;
//                if (count % PROGRESS_INCREMENT == 0) {
//                    logger.info("Processing " + (int)((float)count/(float)numEntity*100) + "% ...");
//                }
//
//                long entityID1 = entityIterator.nextLong();
//                if (!handler.validateSource(entityID1)) {
//                    continue; // skip if prediction_source_filter is not met.
//                }
//
//                if (!rebuild) {
//                    // attention: only do this with database. File doesn't support rebuild for now.
//                    deleteBatch.put(recommenderName, entityID1, updatedTimestamp);
//                }
//                for (RecommendedEntity entity : handler.getRecommendedEntity(entityID1)) {
//                    long entityID2 = entity.getEntityID();
//                    double score = entity.getValue();
//                    if (!handler.validateTarget(entityID2) || !handler.validateScore(score)) {
//                        continue;
//                    }
//
//                    insertBatch.put(recommenderName, entityID1, entityID2, score, updatedTimestamp);
//                    resultCount ++;
//                }
//            }
//            deleteBatch.accomplish();
//            insertBatch.accomplish();

        // this is to compute with multi-threading.
        logger.info("Start computing recommendation and saving results in parallel using threads: " + threadNumber);

        //ThreadPoolExecutor executorService = new ThreadPoolExecutor(getThreadsNumber(), getThreadsNumber(), 10, TimeUnit.SECONDS, new LinkedBlockingQueue<Runnable>());
        ExecutorService executorService = Executors.newFixedThreadPool(threadNumber);
        List<Future<Long>> futuresList = new ArrayList<Future<Long>>();

        while (entityIterator.hasNext()) {
            long entityID1 = entityIterator.nextLong();
            if (!handler.validateSource(entityID1)) {
                continue;
            }

            Callable<Long> callable = null;
            callable = new RecommenderCallable(entityID1, rebuild, deleteBatch, insertBatch, handler);
            Future<Long> future = executorService.submit(callable);
            futuresList.add(future);
        }

        logger.info("Successfully added " + futuresList.size() + " entities. Waiting for processing.");
        executorService.shutdown();
        count = 0;
        for (Future<Long> future : futuresList) {
            count ++;
            if (count % PROGRESS_INCREMENT == 0) {
                logger.info("Processed " + (int)((float)count/(float)futuresList.size()*100) + "% ...");
            }
            try {
                resultCount += future.get();
            } catch (InterruptedException e) {
                throw new DRuntimeException(e);
            } catch (ExecutionException e) {
                throw new DRuntimeException(e);
            }
        }

        try {
            executorService.awaitTermination(Integer.MAX_VALUE, TimeUnit.DAYS);
        } catch (InterruptedException e) {
            throw new DRuntimeException(e);
        }

        assert executorService.isTerminated();
        deleteBatch.accomplish();
        insertBatch.accomplish();


        logger.info("Computing done. Records#: " + resultCount);

        try {
            logger.finest("Waiting for delete batch to be done.");
            deleteBatch.join();
            logger.finest("Waiting for insert batch to be done.");
            insertBatch.join();
        } catch (InterruptedException e) {
            throw new DConnectionException(e);
        }
        // attention: commit here. if there's any unexpected errors, the transaction will be rollback by DBMS
        connection.commit();
        connection.close();
        logger.info("Done saving data.");

        return resultCount;
    }


    protected abstract class EntityHandler {
        public abstract List<RecommendedEntity> getRecommendedEntity(long entityID) throws TasteException;
        public List<RecommendedEntity> wrap(List<RecommendedItem> original) {
            List<RecommendedEntity> wrapped = new ArrayList<RecommendedEntity>();
            for (RecommendedItem item : original) {
                wrapped.add(new RecommendedEntity(item.getItemID(), item.getValue()));
            }
            return wrapped;
        }
        public boolean validateSource(long entityID) {
            return true;  // True if input entity is accepted
        }
        public boolean validateTarget(long entityID) {
            return true;
        }
        public boolean validateScore(double score) {
            return true;
        }
        protected boolean validateSimilarityScore(double score) {
            if (similarityScoreFilter != null && similarityScoreFilter.startsWith(">")) {
                double min = Double.parseDouble(similarityScoreFilter.substring(1));
                return score > min;
            } else {
                return true;
            }
        }
    }

    protected EntityHandler defaultPredictionHandler = new EntityHandler() {
        @Override
        public List<RecommendedEntity> getRecommendedEntity(long entityID) throws TasteException {
            return wrap( recommender.recommend(entityID, maxKeep) );
        }
        @Override
        public boolean validateSource(long entityID) {
            if (predictionSourceFilter != null && predictionSourceFilter.startsWith("<")) {
                long watermark = Long.parseLong(predictionSourceFilter.substring(1));
                return entityID < watermark;
            } else {
                return true;
            }
        }
        @Override
        public boolean validateScore(double score) {
            if (predictionScoreFilter != null && predictionScoreFilter.startsWith(">")) {
                double min = Double.parseDouble(predictionScoreFilter.substring(1));
                return score > min;
            } else {
                return true;
            }
        }
    };

    protected EntityHandler defaultItemSimilarityHandler = new EntityHandler() {
        @Override
        public List<RecommendedEntity> getRecommendedEntity(long entityID) throws TasteException {
            return wrap(((ItemBasedRecommender) recommender).mostSimilarItems(entityID, maxKeep));
        }
        @Override
        public boolean validateScore(double score) {
            return validateSimilarityScore(score);
        }
    };


    private class RecommenderCallable implements Callable<Long> {
        private long entityID1;
        private boolean rebuild;
        private DDatabase.AsyncBatchRunner deleteBatch;
        private DDatabase.AsyncBatchRunner insertBatch;
        private EntityHandler handler;

        public RecommenderCallable(long entityID1, boolean rebuild, DDatabase.AsyncBatchRunner deleteBatch, DDatabase.AsyncBatchRunner insertBatch, EntityHandler handler) {
            this.entityID1 = entityID1;
            this.rebuild = rebuild;
            this.deleteBatch = deleteBatch;
            this.insertBatch = insertBatch;
            this.handler = handler;
        }

        @Override
        public Long call() {
            logger.finest("Computing for entity: " + entityID1);
            long resultCount = 0;
            if (!rebuild) {
                deleteBatch.put(recommenderName, entityID1, updatedTimestamp);
            }
            try {
                for (RecommendedEntity entity : handler.getRecommendedEntity(entityID1)) {
                    long entityID2 = entity.getEntityID();
                    double score = entity.getValue();
                    if (!handler.validateTarget(entityID2) || !handler.validateScore(score)) {
                        continue;
                    }

                    insertBatch.put(recommenderName, entityID1, entityID2, score, updatedTimestamp);
                    resultCount ++;
                }
            } catch (TasteException e) {
                logger.warning("Cannot compute recommendation for entity: " + entityID1);
            }
            return resultCount;
        }
    }


    ////////////// sub classes need to implement this 2 methods.

    abstract protected void computeSimilarity() throws TasteException, DConnectionException, SQLException;

    abstract protected void computePrediction() throws TasteException, DConnectionException, SQLException;

}
