package org.drupal.project.recommender.algorithm;

import org.apache.mahout.cf.taste.common.TasteException;
import org.apache.mahout.cf.taste.impl.common.LongPrimitiveArrayIterator;
import org.apache.mahout.cf.taste.impl.similarity.GenericItemSimilarity;
import org.drupal.project.computing.DConfigException;
import org.drupal.project.computing.DConnectionException;
import org.drupal.project.computing.DRuntimeException;

import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

/**
 * Item2Item incremental version.
 */
public class Item2ItemIncrement extends Item2Item {

    private String baseRecommenderName;

    long lastUpdated;

    @Override
    public void parseConfig(Map<String, Object> params) throws DConfigException {
        super.parseConfig(params);
        // parameters needed for incremental algorithms.
        if (params.containsKey("base_recommender_name")) {
            baseRecommenderName = (String) params.get("base_recommender_name");
        } else {
            throw new DConfigException("The parameter base_recommender_name is required for Item2ItemIncrement algorithm.");
        }
    }

    @Override
    public void run() {

        if (fieldTimestamp == null) {
            throw new DRuntimeException("To run incremental algorithm, you need to specify the timestamp field");
        }
        // max(updated) should be the same for both similarity and prediction.
        Object o = null;
        try {
            o = database.queryValue("SELECT MAX(updated) FROM {recommender_similarity} WHERE recommender = ?", baseRecommenderName);
        } catch (DConnectionException e) {
            throw new DRuntimeException(e);
        }
        if (o == null) {
            lastUpdated = 0;
        } else {
            lastUpdated = (Integer)o;  // very strange, sometimes integer, sometimes long.
        }

        // we basically emulate the app w/ base_id so that data saved using the base app_id.
        recommenderName = baseRecommenderName;
        super.run();
    }

    @Override
    protected void computeSimilarity() throws TasteException, SQLException, DConnectionException {
        List<Map<String, Object>> results = database.query("SELECT DISTINCT " + fieldItem + " FROM " + table + " WHERE " + fieldTimestamp + ">?", lastUpdated);
        if (results.size() == 0) {
            logger.info("No updated items.");
            return;
        }
        long[] updatedItemIDs = new long[results.size()];
        int index = 0;
        for (Map<String, Object> record : results) {
            updatedItemIDs[index] = (Long)record.get(fieldItem);
            index ++;
        }
        numSimilarity = genericComputeSave(defaultItemSimilarityHandler, new LongPrimitiveArrayIterator(updatedItemIDs), updatedItemIDs.length, "similarity", false);
    }

    @Override
    protected void computePrediction() throws TasteException, SQLException, DConnectionException {
        // compute new users
        List<Map<String, Object>> results = database.query("SELECT DISTINCT " + fieldUser + " FROM " + table + " WHERE " + fieldTimestamp + ">?", lastUpdated);
        if (results.size() == 0) {
            logger.info("No updated users.");
            return;
        }
        long[] updatedUserIDs = new long[results.size()];
        int index = 0;
        for (Map<String, Object> record : results) {
            updatedUserIDs[index] = (Long)record.get(fieldUser);
            index ++;
        }

        // <del>TODO: [#1188294] for multi-dbms support.</del> we'll load data manually.
        // reload similarity data from database rather than compute again.
        List<GenericItemSimilarity.ItemItemSimilarity> precomputedSimilarity = new ArrayList<GenericItemSimilarity.ItemItemSimilarity>();
        List<Map<String, Object>> results1 = database.query("SELECT source_id, target_id, score FROM {recommender_similarity} WHERE recommender=?", recommenderName);
        for (Map<String, Object> record : results1) {
            precomputedSimilarity.add(new GenericItemSimilarity.ItemItemSimilarity((Long)record.get("source_id"), (Long)record.get("target_id"), (Float)record.get("score")));
        }

        // override itemSimilarity to use existing similarity, and reload recommender
        itemSimilarity = new GenericItemSimilarity(precomputedSimilarity);
        initRecommender();
        numPrediction = genericComputeSave(defaultPredictionHandler, new LongPrimitiveArrayIterator(updatedUserIDs), updatedUserIDs.length, "prediction", false);
    }

}
